#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>	/* inet_aton,inet_ntoa */
#include <unistd.h>	/* read,write,close */
#include <string.h>	/* memset */
#include <stdlib.h>	/* malloc */
#include <stdio.h>	/* FILE,fprintf */

#define FILE_READ_MAX_SIZE 655000
#define FILE_ONE_PAGE_SIZE 64000

#define SDRAM_MAX_LENGTH 640000
#define TOTAL_SETTINGS_SIZE 4096

#define TCP_DOWNLOAD_SIZE 128
#define TCP_DOWNLOAD_PORT 0x2223 // 8739

#define DATA_HINT "dh"
#define SETTINGS_HINT "sh"

/*
 * Download protocol (from PalmToolDlg.cpp):
 * Step1. PC make a TCP connection to Palm1's page0 program
 * Step2. PC sends 2 bytes to Palm1 for data "dh" or settings "sh" hints
 *                Palm1 sends back 128 byte of data
 *        Repeat step2 until the whole data has been received
 * Step3. PC close the TCP connection indicate the download is over
 */
int download(int sd, char *buffer, size_t length, const char *hint) {
	fprintf(stdout, "Downloading... ");
	// receiving data
	while (length > 0) {
		write(sd, hint, strlen(hint));
		read(sd, buffer, TCP_DOWNLOAD_SIZE);
		// FIXME: should check for errors!
		fprintf(stdout, "%10d\b\b\b\b\b\b\b\b\b\b", length);
		buffer += TCP_DOWNLOAD_SIZE;
		length -= TCP_DOWNLOAD_SIZE;
	}
	fprintf(stdout, "Done.     \n");
	return 0;
}

int main(int argc, const char *argv[]) {
	const char *progname, *filename, *extension, *hint, *ipaddr;
	FILE *filehandle;
	size_t length, length_padded;
	char *buffer;
	struct in_addr sin_addr;
	int sd;
	struct sockaddr_in sin;

	// check arguments count
	progname = argv[0];
	if (argc != 3) {
		fprintf(stderr, "Syntax: %s <filename> <IPv4-address>\n", progname);
		return 1;
	}
	filename = argv[1];
	ipaddr = argv[2];

	// check first argument (filename)
	filehandle = fopen(filename, "wb");
	if (filehandle == NULL) {
		fprintf(stderr, "Unable to open '%s' !\n", filename);
		return 2;
	}
	fprintf(stdout, "File name: %s\n", filename);

	// check file type (by extension, for now...)
	extension = strrchr(filename, '.');
	if (extension && !strcasecmp(extension, ".dh")) {
		hint = DATA_HINT;
		length = SDRAM_MAX_LENGTH; // < FILE_READ_MAX_SIZE
	} else if (extension && !strcasecmp(extension, ".sh")) {
		hint = SETTINGS_HINT;
		length = TOTAL_SETTINGS_SIZE; // < FILE_ONE_PAGE_SIZE
	} else {
		fprintf(stderr, "Unknown file type!\n");
		return 2;
	}

	// normalize file size
	length_padded = length + TCP_DOWNLOAD_SIZE*2 - 1;
	length_padded /= TCP_DOWNLOAD_SIZE*2;
	length_padded *= TCP_DOWNLOAD_SIZE*2;

	// prepare to load data into memory
	fprintf(stdout, "File length: %d bytes", length);
	if (length_padded > length) {
		fprintf(stdout, " (padding to %d bytes)", length_padded);
	}
	fprintf(stdout, "\n");
	buffer = malloc(length_padded);
	if (buffer == NULL) {
		fprintf(stderr, "Unable to allocate enough (%d bytes) memory!\n", length_padded);
		return 4;
	}

	// check second argument (IPv4 address)
	if (!inet_aton(ipaddr, &sin_addr)) {
		fprintf(stderr, "'%s' is an invalid IPv4 address!\n", ipaddr);
		return 6;
	}
	fprintf(stdout, "IPv4 address: %s\n", inet_ntoa(sin_addr));

	// connect to the host
	sd = socket(PF_INET, SOCK_STREAM, 0);
	if (sd < 0) {
		fprintf(stderr, "Unable to create a socket!\n");
		return 7;
	}
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(TCP_DOWNLOAD_PORT);
	memcpy(&sin.sin_addr, &sin_addr, sizeof(sin.sin_addr));
	if (connect(sd, (struct sockaddr *)&sin, sizeof(sin))) {
		fprintf(stderr, "Unable to connect to host!\n");
		return 8;
	}

	// do the upgrade process
	setbuf(stdout, NULL);
	if (download(sd, buffer, length_padded, hint)) return 9;

	// save to file
	//if (length < length_padded) {
	//	memset(buffer + length, 0, length_padded - length);
	//}
	if (fwrite(buffer, length, 1, filehandle) != 1) { // XXX: length_padded?
		fprintf(stderr, "Unable to write the whole file content!\n");
		return 5;
	}
	fclose(filehandle);

	// cleaning
	close(sd);
	return 0;
}
