#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>	/* inet_aton,inet_ntoa */
#include <signal.h>	/* signal,SIG* */
#include <errno.h>	/* errno,E* */
#include <unistd.h>	/* read,write,close */
#include <string.h>	/* strcmp,memset */
#include <stdio.h>	/* FILE,fprintf */

#define WAIT_TIME_MANY    3 // seconds
#define WAIT_TIME_ONLYONE 1 // seconds

#define UDP_DEBUG_SRC_PORT 1025

#define REPORT_REQUEST_STRING "Hello, world!"
#define REPORT_DST_IPADDR "224.1.2.5"
#define REPORT_SRC_PORT 61030

#define REPORT_IP_POS		0	// 0-3 IP
#define REPORT_MAC_POS		4	// 4-9 MAC
#define REPORT_VERSION_POS	10	// 10-39 Version
#define REPORT_NUMBER_POS	40	// 40-59 Phone Number
#define REPORT_TYPE_POS		60	// 60 ip type
#define REPORT_LOGON_POS	61	// 61 Logon status
#define REPORT_CALL_POS		62	// 62 Call status
#define REPORT_PROTOCOL_POS	63	// 63 Protocol
#define REPORT_TOTAL_LEN	64

static const char *report_type[] = {
  "static", "dhcp", "pppoe", "modem"
};

static const char *report_protocol[] = {
  "h323", "sip", "mgcp", "n2p", "iax2", "wp", "epn", "others"
};

void human_report(const unsigned char *buffer) {
	const unsigned char *p;
	fprintf(stdout, "================\n");
	p = buffer + REPORT_IP_POS;
	fprintf(stdout, "IPv4 address...: %d.%d.%d.%d\n", p[0],p[1],p[2],p[3]);
	p = buffer + REPORT_MAC_POS;
	fprintf(stdout, "MAC address....: %02x-%02x-%02x-%02x-%02x-%02x\n", p[0],p[1],p[2],p[3],p[4],p[5]);
	p = buffer + REPORT_VERSION_POS;
	fprintf(stdout, "Version........: %-30.30s\n", p);
	p = buffer + REPORT_NUMBER_POS;
	fprintf(stdout, "Phone Number...: %-20.20s\n", p);
	p = buffer + REPORT_TYPE_POS;
	fprintf(stdout, "IP Type........: %d (%s)\n", p[0], (p[0] < sizeof(report_type)/sizeof(char*)) ? report_type[p[0]] : "unknown");
	p = buffer + REPORT_LOGON_POS;
	fprintf(stdout, "Logged on......: %d (%s)\n", p[0], (p[0] == 1 ? "Yes" : "No"));
	p = buffer + REPORT_CALL_POS;
	fprintf(stdout, "Busy...........: %d (%s)\n", p[0], (p[0] == 1 ? "Yes" : "No"));
	p = buffer + REPORT_PROTOCOL_POS;
	fprintf(stdout, "Protocol.......: %d (%s)\n", p[0], (p[0] < sizeof(report_protocol)/sizeof(char*)) ? report_protocol[p[0]] : "unknown");
}

void machine_report(const unsigned char *buffer) {
	const unsigned char *p;
	p = buffer + REPORT_IP_POS;
	fprintf(stdout, "%d.%d.%d.%d:", p[0],p[1],p[2],p[3]);
	p = buffer + REPORT_MAC_POS;
	fprintf(stdout, "%02x-%02x-%02x-%02x-%02x-%02x:", p[0],p[1],p[2],p[3],p[4],p[5]);
	p = buffer + REPORT_VERSION_POS;
	fprintf(stdout, "%.30s:", p);
	p = buffer + REPORT_NUMBER_POS;
	fprintf(stdout, "%.20s:", p);
	p = buffer + REPORT_TYPE_POS;
	fprintf(stdout, "%s:", (p[0] < sizeof(report_type)/sizeof(char*)) ? report_type[p[0]] : "unknown");
	p = buffer + REPORT_LOGON_POS;
	fprintf(stdout, "%s:", (p[0] == 1 ? "Yes" : "No"));
	p = buffer + REPORT_CALL_POS;
	fprintf(stdout, "%s:", (p[0] == 1 ? "Yes" : "No"));
	p = buffer + REPORT_PROTOCOL_POS;
	fprintf(stdout, "%s\n", (p[0] < sizeof(report_protocol)/sizeof(char*)) ? report_protocol[p[0]] : "unknown");
}

char alarm_signal = 0;
void signal_handler(int signal_number) {
	if (signal_number == SIGALRM) alarm_signal = 1;
}

int main(int argc, const char *argv[]) {
	const char *progname, *ipaddr;
	struct in_addr sin_addr;
	int only_one, answers_count;
	int sd;
	struct sockaddr_in sin, sinreq;
	void (*report_function)(const unsigned char *) = machine_report;

	// check arguments count
	progname = argv[0];
	if (argc >= 2 && !strcmp(argv[1], "vertical")) {
		report_function = human_report;
		argc--; argv++;
	}
	if (argc != 1 && argc != 2) {
		fprintf(stderr, "Syntax: %s [vertical] [<IPv4-address>]\n", progname);
		return 1;
	}
	setbuf(stdout, NULL);

	// check second argument (IPv4 address)
	if (argc == 2) {
		ipaddr = argv[1];
		only_one = 1;
	} else {
		ipaddr = REPORT_DST_IPADDR;
		only_one = 0;
	}
	if (!inet_aton(ipaddr, &sin_addr)) {
		fprintf(stderr, "'%s' is an invalid IPv4 address!\n", ipaddr);
		return 2;
	}

	// prepare receiving answers (bind report port)
	sd = socket(PF_INET, SOCK_DGRAM, 0);
	if (sd < 0) {
		fprintf(stderr, "Unable to create a socket!\n");
		return 3;
	}
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(REPORT_SRC_PORT);
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	if (bind(sd, (struct sockaddr *)&sin, sizeof(sin))) {
		fprintf(stderr, "Unable to bind!\n");
		return 4;
	}

	// send report request
	memset(&sinreq, 0, sizeof(sinreq));
	sinreq.sin_family = AF_INET;
	sinreq.sin_port = htons(UDP_DEBUG_SRC_PORT);
	memcpy(&sinreq.sin_addr, &sin_addr, sizeof(sinreq.sin_addr));
	if (sendto(sd, REPORT_REQUEST_STRING, sizeof(REPORT_REQUEST_STRING), 0, (struct sockaddr *)&sinreq, sizeof(sinreq)) < 0) {
		fprintf(stderr, "Error sending UDP request!\n");
		return 5;
	}

	// wait for answers
	if (!only_one) {
		fprintf(stdout, "#ipv4-address:mac-address:version:phone-number:connexion:logged:busy:protocol\n");
	}
	signal(SIGPIPE, SIG_IGN);
	signal(SIGALRM, signal_handler);
	alarm(only_one ? WAIT_TIME_ONLYONE : WAIT_TIME_MANY);
	answers_count = 0;
	while (!(only_one && (answers_count == 1)) && !alarm_signal) {
		unsigned char buffer[REPORT_TOTAL_LEN];
		int size = recv(sd, buffer, sizeof(buffer), MSG_DONTWAIT);
		if (size == -1 && errno == EAGAIN) {
			usleep(200);
		} else if (size == REPORT_TOTAL_LEN) {
			(*report_function)(buffer);
			answers_count++;
		} else {
			fprintf(stderr, "Warning: received an invalid packet size of %d bytes!\n", size);
		}
	}

	// cleaning
	close(sd);
	return (only_one && (answers_count != 1)) ? -1 : 0;
}
