#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>	/* inet_aton,inet_ntoa */
#include <unistd.h>	/* read,write,close */
#include <string.h>	/* memset */
#include <stdlib.h>	/* malloc */
#include <stdio.h>	/* FILE,fprintf */

#define FILE_FULL_PAGE_SIZE 65536

#define FIRMWARE_SIZE (FILE_FULL_PAGE_SIZE*15)
#define MAX_PCMRING_SIZE 15360
#define SDRAM_MAX_LENGTH 640000
#define TOTAL_SETTINGS_SIZE 4096

#define TCP_UPLOAD_SIZE 128
#define TCP_UPLOAD_PORT 0x2223 // 8739

#define OPTIONS_UPLOAD_PAGE 128
#define PROGRAM_UPLOAD_PAGE 129
#define RING_UPLOAD_PAGE 130
#define MAP_UPLOAD_PAGE 132
#define ALTERNATE_SETTING_UPLOAD_PAGE 133

/*
 * Upload protocol (from PalmToolDlg.cpp):
 * Step1. PC make a TCP connection to Palm1's page0 program
 * Step2. PC sends 128 bytes package data to Palm1
 *                Palm1 send back 1 byte of "0xff"
 *        Repeat step2 until the whole data has been sent
 * Step3. PC sends 1 byte data indicate which page to write to
 *                Palm1 send back 1 byte of "0xff"
 * Step4. PC close the TCP connection indicate the update progress is over
 */
int upload(int sd, const char *buffer, size_t length, int upload_page) {
	unsigned char result;
	fprintf(stdout, "Uploading... ");
	// sending data
	while (length > 0) {
		write(sd, buffer, TCP_UPLOAD_SIZE);
		read(sd, &result, 1);
		if (result != 0xff) {
			fprintf(stderr, "Warning: unexpected read result (%02X) !\n", result);
			fprintf(stdout, "Aborting.\n");
			return 1;
		}
		fprintf(stdout, "%10d\b\b\b\b\b\b\b\b\b\b", length);
		buffer += TCP_UPLOAD_SIZE;
		length -= TCP_UPLOAD_SIZE;
	}
	// sending page number
	result = upload_page;
	write(sd, &result, 1);
	read(sd, &result, 1);
	if (result != 0xff) {
		fprintf(stderr, "Warning: unexpected read result (%02X) !\n", result);
		fprintf(stdout, "Aborting.\n");
		return 1;
	}
	fprintf(stdout, "Done.     \n");
	return 0;
}

int main(int argc, const char *argv[]) {
	const char *progname, *filename, *extension, *ipaddr;
	FILE *filehandle;
	int upload_page;
	size_t length, length_padded;
	char *buffer;
	struct in_addr sin_addr;
	int sd;
	struct sockaddr_in sin;

	// check arguments count
	progname = argv[0];
	if (argc != 3) {
		fprintf(stderr, "Syntax: %s <filename> <IPv4-address>\n", progname);
		fprintf(stderr, "  where <filename> is one of:\n");
		fprintf(stderr, "        <name>.bin : firmware\n");
		fprintf(stderr, "        <name>.dat : ring\n");
		fprintf(stderr, "        <name>.sh  : options\n");
		fprintf(stderr, "        map.txt    : digit map\n");
		fprintf(stderr, "        stdram.dat : \n");
		return 1;
	}
	filename = argv[1];
	ipaddr = argv[2];

	// check first argument (filename)
	filehandle = fopen(filename, "rb");
	if (filehandle == NULL) {
		fprintf(stderr, "Unable to open '%s' !\n", filename);
		return 2;
	}
	fprintf(stdout, "File name: %s\n", filename);

	// check file type (by extension, for now...)
	extension = strrchr(filename, '.');
	if (!strcasecmp(filename, "sdram.dat")) {
		// FIXME: upload_page = ;
		fprintf(stderr, "Not implemented yet!\n");
		return 2;
	} else if (!strcasecmp(filename, "map.txt")) {
		upload_page = MAP_UPLOAD_PAGE;
	} else if (extension && !strcasecmp(extension, ".sh")) {
		upload_page = OPTIONS_UPLOAD_PAGE;
	} else if (extension && !strcasecmp(extension, ".bin")) {
		upload_page = PROGRAM_UPLOAD_PAGE;
	} else if (extension && !strcasecmp(extension, ".dat")) {
		upload_page = RING_UPLOAD_PAGE;
	} else {
		fprintf(stderr, "Unknown file type!\n");
		return 2;
	}

	// get file size
	fseek(filehandle, 0L, SEEK_END);
	length = ftell(filehandle);
	length_padded = length + TCP_UPLOAD_SIZE*2 - 1;
	length_padded /= TCP_UPLOAD_SIZE*2;
	length_padded *= TCP_UPLOAD_SIZE*2;

	// check file size
	if ((upload_page == PROGRAM_UPLOAD_PAGE) && (length_padded != FIRMWARE_SIZE)) {
		fprintf(stderr, "Invalid file size (%d != %d) !\n", length_padded, FIRMWARE_SIZE);
		return 3;
	} else if ((upload_page == RING_UPLOAD_PAGE) && (length_padded > MAX_PCMRING_SIZE)) {
		fprintf(stderr, "Invalid file size (%d > %d) !\n", length_padded, MAX_PCMRING_SIZE);
		return 3;
	} else if ((upload_page == OPTIONS_UPLOAD_PAGE) && (length_padded != TOTAL_SETTINGS_SIZE)) {
		fprintf(stderr, "Invalid file size (%d > %d) !\n", length_padded, MAX_PCMRING_SIZE);
		return 3;
	}

	// load data into memory
	fprintf(stdout, "File length: %d bytes", length);
	if (length_padded > length) {
		fprintf(stdout, " (padding to %d bytes)", length_padded);
	}
	fprintf(stdout, "\n");
	buffer = malloc(length_padded);
	if (buffer == NULL) {
		fprintf(stderr, "Unable to allocate enough (%d bytes) memory!\n", length_padded);
		return 4;
	}
	fseek(filehandle, 0L, SEEK_SET);
	if (fread(buffer, length, 1, filehandle) != 1) {
		fprintf(stderr, "Unable to read the whole file content!\n");
		return 5;
	}
	fclose(filehandle);
	if (length < length_padded) {
		memset(buffer + length, 0, length_padded - length);
	}

	// check second argument (IPv4 address)
	if (!inet_aton(ipaddr, &sin_addr)) {
		fprintf(stderr, "'%s' is an invalid IPv4 address!\n", ipaddr);
		return 6;
	}
	fprintf(stdout, "IPv4 address: %s\n", inet_ntoa(sin_addr));

	// connect to the host
	sd = socket(PF_INET, SOCK_STREAM, 0);
	if (sd < 0) {
		fprintf(stderr, "Unable to create a socket!\n");
		return 7;
	}
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(TCP_UPLOAD_PORT);
	memcpy(&sin.sin_addr, &sin_addr, sizeof(sin.sin_addr));
	if (connect(sd, (struct sockaddr *)&sin, sizeof(sin))) {
		fprintf(stderr, "Unable to connect to host!\n");
		return 8;
	}

	// do the upgrade process
	setbuf(stdout, NULL);
	if (upload(sd, buffer, length_padded, upload_page)) return 9;

	// cleaning
	close(sd);
	return 0;
}
